#ifndef _TASMMATH_H_
#define _TASMMATH_H_

#include "allegro.h"

fixed calcAngle (const fixed & fromX, const fixed & fromY, const fixed & toX, const fixed & toY, const fixed & currAngle);

fixed calcDist (const fixed & fromX, const fixed & fromY, const fixed & toX, const fixed & toY);

fixed angleDist (const fixed & fromAngle, const fixed & toAngle);

fixed fixabs (const fixed & f);

fixed normalizeAngle (const fixed & f);

void setUpDialog (DIALOG *, int (*proc)(int, DIALOG *, int), int x, int y, int w, int h, int fg, int bg, int key, int flags, int d1, int d2, void *dp, void *dp2, void *dp3);
#endif
